package polyglot.ext.polyj.ast;

import polyglot.ext.jl.ast.*;

import polyglot.ext.polyj.types.*;

import polyglot.ast.*;
import polyglot.types.*;
import polyglot.visit.*;
import polyglot.util.*;

import java.util.*;

public class AmbNewArray_c extends NewArray_c implements AmbNewArray
{
    /**
     * The first 'dimension' in the array creation; "B" in "new A[B][C]".
     * ('dimension' since it may not actually be such, but could be
     * a parametric instantiation.)
     */
    protected Receiver param;
    
    /**
     * Create a node representing "baseType[param][dims...]".
     */
    public AmbNewArray_c(Position pos, TypeNode baseType, Receiver param, List dims,
                         int addDims, ArrayInit init) {
        
        super(pos, baseType, dims, addDims, init);
        this.param = param;
    }

    public Receiver param() {
        return this.param;
    }

    public AmbNewArray param(Receiver param) {
        AmbNewArray_c n = (AmbNewArray_c) copy();
        n.param = param;
        return n;
    }

    protected AmbNewArray_c reconstruct(TypeNode baseType, Receiver param, List dims, ArrayInit init) {
        if (baseType != this.baseType 
            || param != this.param
            || ! CollectionUtil.equals(dims, this.dims)
            || init != this.init) {
            AmbNewArray_c n = (AmbNewArray_c) copy();
            n.baseType = baseType;
            n.param = param;
            n.dims = TypedList.copyAndCheck(dims, Expr.class, true);
            n.init = init;
            return n;
        }

        return this;
    }

    public Node visitChildren(NodeVisitor v) {
        TypeNode baseType = (TypeNode) visitChild(this.baseType, v);
        Receiver param = (Receiver) visitChild(this.param, v);
        List dims = visitList(this.dims, v);
        ArrayInit init = (ArrayInit) visitChild(this.init, v);
        return reconstruct(baseType, param, dims, init);
    }

    public String toString() {
        return "new " + baseType + "[" + param + "]...{amb}";
    }

    public Node disambiguate(AmbiguityRemover ar) throws SemanticException {
        PolyJTypeSystem ts = (PolyJTypeSystem) ar.typeSystem();
        PolyJNodeFactory nf = (PolyJNodeFactory) ar.nodeFactory();
        
        if (dims.isEmpty()) {
            throw new InternalCompilerError(position(),
                "Cannot disambiguate ambiguous new array with no " +
                "dimension expressions.");
        }

        Type t = baseType.type();
        
        // Now that "baseType" has been disambiguated, we can tell whether
        // the first dimension is a type or an access.

        if (t instanceof PolyJClassType) {
            // "param" is a parameter.  Instantiate the base type with the
            // parameter and use it as the new base type.            
            PolyJClassType pt = (PolyJClassType) t;

            if (pt.params().size() != 1) {
                //this node shouldn't be ambiguous.
                throw new InternalCompilerError(position(), "An AmbNewArray starting with a parameterized " +
                    "type should have only one parameter to that type.");
            }
            
            ParamNode disambParam = nf.Param(param.position(), param);
            
            List l = new LinkedList();
            l.add(disambParam.parameter());
            
            Type newBaseType = ts.instantiate(baseType.position(), pt, l);

            return nf.NewArray(position(),
                               nf.CanonicalTypeNode(baseType.position(), newBaseType),
                               dims).additionalDims(addDims).init(init);
            
        } else {
            // "param" must be an expression.  Throw it in with the rest of the
            // accesses.
            
            Expr disambParam = (Expr) visitChild(param, ar);
            
            List l = new LinkedList();
            l.add(disambParam);
            l.addAll(dims);
            
            return nf.NewArray(position(), baseType, l).additionalDims(addDims).init(init);
        }
    }
    
    public Node typeCheck(TypeChecker tc) throws SemanticException {
        throw new InternalCompilerError(position(),
            "Cannot type check ambiguous node " + this + ".");
    } 

    public Node exceptionCheck(ExceptionChecker ec) throws SemanticException {
        throw new InternalCompilerError(position(),
            "Cannot exception check ambiguous node " + this + ".");
    } 

    public void prettyPrint(CodeWriter w, PrettyPrinter tr) {
        w.write("new ");
        print(baseType, w, tr);

        w.write("[");
        print(param, w, tr);
        w.write("]");

        for (Iterator i = dims.iterator(); i.hasNext();) {
            Expr e = (Expr) i.next();
            w.write("[");
            printBlock(e, w, tr);
            w.write("]");
        }

        for (int i = 0; i < addDims; i++) {
            w.write("[]");
        }

        if (init != null) {
            w.write(" ");
            print(init, w, tr);
        }
    }

    public void translate(CodeWriter w, Translator tr) {
        throw new InternalCompilerError(position(),
            "Cannot translate ambiguous node " + this + ".");
    }
}
