package polyglot.ext.polyj.ast;

import polyglot.ext.polyj.types.*;

import polyglot.ast.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;

import java.util.*;

public class MethodConstraint_c extends ProcedureConstraint_c implements MethodConstraint {
    protected TypeNode returns;
    
    public MethodConstraint_c( Position pos, Flags flags, TypeNode tn, 
                                   String name, List formals, List exceptions)
    {
        super(pos);
        this.flags = flags;
        this.returns = tn;
        this.name = name;
        this.formals = TypedList.copyAndCheck(formals, TypeNode.class, true);
        this.exceptions = TypedList.copyAndCheck(exceptions, TypeNode.class, true);

        if (tn == null) {
            throw new InternalCompilerError("MethodConstraint: " +
                "method with null return type");
        }
    }

    /**
     * Lazily reconstruct this node.
     */
    protected MethodConstraint reconstruct(TypeNode returns, List formals, List exceptions)
    {
        if(   this.returns != returns 
           || !CollectionUtil.equals(this.formals, formals)
           || !CollectionUtil.equals(this.exceptions, exceptions) ) {
            MethodConstraint_c n = (MethodConstraint_c) copy();
            n.returns = returns;
            n.formals = TypedList.copyAndCheck(formals, TypeNode.class, true);
            n.exceptions = TypedList.copyAndCheck(exceptions, TypeNode.class, true);
            return n;
        } else {
            return this;
        }
    }

    /**
     * Returns the TypeNode representing the return type of this method.
     */
    public TypeNode returnType() 
    {
        return returns;
    }

    public MethodConstraint returnType(TypeNode returnType) {
        MethodConstraint_c n = (MethodConstraint_c) copy();
        n.returns = returnType;
        return n;
    }
    
    public Node visitChildren(NodeVisitor v) 
    {
        TypeNode newReturns = (TypeNode) visitChild(returns, v);
        List newFormals = visitList(formals, v);
        List newExceptions = visitList(exceptions, v);
        return reconstruct(newReturns, newFormals, newExceptions);
    }

    public void prettyPrint(CodeWriter w, PrettyPrinter tr)
    {
        w.begin(0);
        w.write(flags.translate());
        print(returns, w, tr);
        w.write(" " + name + "(");
        w.begin(0);

        for(Iterator i = formals.iterator(); i.hasNext();)
        {
            TypeNode tn = (TypeNode) i.next();
            print(tn, w, tr);
            if(i.hasNext()) {
                w.write(",");
                w.allowBreak(0, " ");
            }
        }
        w.end();
        w.write(")");

        if(!exceptions.isEmpty()) {
            w.allowBreak(6);
            w.write("throws ");
            for(Iterator i = exceptions.iterator(); i.hasNext();) {
                TypeNode tn = (TypeNode) i.next();
                print(tn, w, tr);
                if (i.hasNext()) {
                    w.write(","); w.allowBreak(4, " ");
                }
            }
        }
        w.write(";");
        w.end();
    }

    public void dump(CodeWriter w) 
    {
        w.write("POLYJ METHOD CONSTRAINT");
        w.write(" < " + name + " >");
        w.write(" < " + returns.type().toString() + " > ");
        w.write(" < " + flags.toString() + "> ");
    }

    /** 
     * Convert this into a MethodInstance associated with the given ParamType.
     */
    public ProcedureInstance toInstance(ParamType param) {
        List formalTypes = new ArrayList(formals.size());
        List throwTypes = new ArrayList(exceptions.size());

        for (Iterator it = formals.iterator(); it.hasNext(); ) {
            formalTypes.add((Type)((TypeNode)it.next()).type());
        }

        for (Iterator it = exceptions.iterator(); it.hasNext(); ) {
            throwTypes.add((Type)((TypeNode)it.next()).type());
        }

        Flags af = (Flags) flags.Public().Abstract();
        
        MethodInstance mti = param.typeSystem().methodInstance(position(),
                               param, af, returns.type(), name, formalTypes, throwTypes);
    
        return mti;
    }
}
