package polyglot.ext.polyj.ast;

import polyglot.ast.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;

import polyglot.ext.polyj.visit.*;
import polyglot.ext.polyj.util.*;
import polyglot.ext.polyj.types.*;
import polyglot.ext.polyj.extension.*;

import polyglot.ext.jl.ast.*;

import java.util.*;
import java.io.*;

public class PolyJClassDecl_c extends ClassDecl_c implements PolyJClassDecl {
    protected List params;
    protected List constraints;

    /**
     * Holds the short name of the adapter interface for this class.
     * Set upon first entering translate().
     */
    protected String adIntName;

    public PolyJClassDecl_c( Position pos, Flags flags, String name,
                            List params, List constraints, TypeNode superClass,
                            List interfaces, ClassBody body) {
        super(pos, flags, name, superClass, interfaces, body);
        this.params = TypedList.copyAndCheck(params, TypeNode.class, true);
        this.constraints = TypedList.copyAndCheck(constraints,
                                                  ClassConstraint.class, true);
    }

    protected ClassDecl reconstruct(List params, List constraints,
                                    TypeNode superClass, List interfaces,
                                    ClassBody body) {

        if (! CollectionUtil.equals(this.params, params) ||
            ! CollectionUtil.equals(this.constraints, constraints) ||
              this.superClass != superClass ||
            ! CollectionUtil.equals(this.interfaces, interfaces) ||
              this.body != body) {

            PolyJClassDecl_c n = (PolyJClassDecl_c)
                super.reconstruct(superClass, interfaces, body);
            n.params = TypedList.copyAndCheck(params, TypeNode.class, true);
            n.constraints = TypedList.copyAndCheck(constraints,
                                                   ClassConstraint.class, true);
            return n;
        } else {
            return this;
        }
    }

    /**
     * Visit the children of this node.
     *
     * @pre Requires that <code>superClass.visit</code> returns an object of
     *  type <code>TypeNode</code> and that the <code>visit</code> method of 
     *  each interface and member returns an object of types 
     *  <code>TypeNode</code> and <code>ClassMember</code> respectively.
     * @post Returns <code>this</code> if no changes are made, otherwise a copy
     *  is made and returned.   
     */
    public Node visitChildren(NodeVisitor v) {
        TypeNode newSuperClass = (TypeNode) visitChild(superClass, v);
        List newParams = visitList(params, v);
        List newConstraints = visitList(constraints, v);
        List newInterfaces = visitList(interfaces, v);
        ClassBody newBody = (ClassBody) visitChild(body, v);
        return reconstruct(newParams, newConstraints, newSuperClass,
                           newInterfaces, newBody);
    }

    public List params() {
        return params;
    }

    public PolyJClassDecl params(List params) {
        PolyJClassDecl_c n = (PolyJClassDecl_c) copy();
        n.params = TypedList.copyAndCheck(params, TypeNode.class, true);
        return n;
    }

    public List constraints() {
        return constraints;
    }

    public PolyJClassDecl constraints(List constraints) {
        PolyJClassDecl_c n = (PolyJClassDecl_c) copy();
        n.constraints = TypedList.copyAndCheck(constraints,
                                               ClassConstraint.class, true);
        return n;
    }

    public void dump(CodeWriter w) {
        w.write( "POLYJCLASS < " + name + " > ");
        w.write( " < " + flags.toString() + " > ");
        w.write( "PARAMS:< ");
        for (Iterator i = params.iterator(); i.hasNext(); ) {
            TypeNode p = (TypeNode)i.next();
            p.dump(w);
        }
        w.write( ">");
        w.write( "CONSTRAINTS:< ");
        for (Iterator i = constraints.iterator(); i.hasNext(); ) {
            ClassConstraint c = (ClassConstraint)i.next();
            c.dump(w);
        }
        w.write( ">");
    }

    public Node buildTypes(TypeBuilder tb) throws SemanticException {
        // We want to create the ParamType for each param.
        // We delay creating the constraints until the type disambiguation pass.
        PolyJClassDecl_c n = (PolyJClassDecl_c) super.buildTypes(tb);
        n.addParams();
        return n;
    }

    protected void addParams() throws SemanticException {
        PolyJClassType ct = (PolyJClassType) this.type;

        for (Iterator i = params.iterator(); i.hasNext(); ) {
            AmbTypeNode p = (AmbTypeNode) i.next();
            ct.addParam(p.name());
        }
    }

    protected void addConstraints(List constraints) throws SemanticException {
        PolyJClassType ct = (PolyJClassType) this.type;

        for (Iterator i = constraints.iterator(); i.hasNext(); ) {
            ClassConstraint c = (ClassConstraint) i.next();
            ct.addConstraint(c);
        }
    }

    public Context enterScope(Context c) {
        PolyJClassType myType = (PolyJClassType) type;

        c = super.enterScope(c);

        // NJN: This seems redundant since pushClass also does this, but
        // removing causes the params not to be found.
        if (myType != null) {
            return ((PolyJContext) c).pushParams(myType.params().iterator());
        }

        return c;
    }

    public Node addMembers(AddMemberVisitor ar) throws SemanticException {
        PolyJClassDecl_c newNode = (PolyJClassDecl_c) super.addMembers(ar);

        PolyJClassType newType = (PolyJClassType) newNode.type();
        PolyJContext c = (PolyJContext) ar.context();
        PolyJTypeSystem ts = (PolyJTypeSystem) ar.typeSystem();

        // Disambiguate the constraints
        List newConstraints = newNode.constraints;

        // Add the constraints to this class, which amounts
        // to creating MemberInstances for all the members it declares
        // and adding them to the appropriate ParamType.
        // It's important that this occur *after* the constraints' types
        // are disambiguated, because we lose the ability to disambiguate
        // them once they're added.  (See ProcedureConstraint.toInstance,
        // where everything but the type gets lost.)
        newNode.addConstraints(newConstraints);
        for (Iterator parIt = newType.params().iterator(); parIt.hasNext(); ) {
            ParamType paramType = (ParamType) parIt.next();

            for (Iterator constrIt = paramType.constructors().iterator(); constrIt.hasNext(); ) {
                ConstructorInstance ci = (ConstructorInstance) constrIt.next();

                // ts.addWhereBinding(paramType, ci, c);
            }

            for (Iterator methIt = paramType.methods().iterator(); methIt.hasNext(); ) {
                MethodInstance mi = (MethodInstance) methIt.next();

                // Populate the wherebinding tables at this point.
                //   we need the Context because we need to
                //   establish a correspondence between the where-bindings
                //   of this class with the wherebindings of its supertype.
                // ts.addWhereBinding(paramType, mi, c);
            }
        }

        return newNode;
    }

    public void prettyPrint(CodeWriter w, PrettyPrinter tr) {
            if (flags.isInterface()) {
                w.write(flags.clearInterface().clearAbstract().translate());
            }
            else {
                w.write(flags.translate());
            }

            if (flags.isInterface()) {
                w.write("interface ");
            }
            else {
                w.write("class ");
            }

            w.write(name);

            if (! params.isEmpty()) {
                w.write("[");
                for (Iterator i = params.iterator(); i.hasNext(); ) {
                    TypeNode tn = (TypeNode) i.next();
                    print(tn, w, tr);
                    if (i.hasNext()) {
                        w.write(", ");
                    }
                }
                w.write("]");
            }
    
            if (! constraints.isEmpty()) {
                w.begin(0);
                for (Iterator i = constraints.iterator(); i.hasNext(); ) {
                    ClassConstraint cc = (ClassConstraint) i.next();
                    w.write(" ");
                    printBlock(cc, w, tr);
                    if (i.hasNext()) {
                        w.allowBreak(0, " ");
                    }
                }
                w.end();
            }
    
            if (superClass() != null) {
                    w.write(" extends ");
                    print(superClass(), w, tr);
            }

            if (! interfaces.isEmpty()) {
                    if (flags.isInterface()) {
                            w.write(" extends ");
                    }
                    else {
                            w.write(" implements ");
                    }

                    for (Iterator i = interfaces().iterator(); i.hasNext(); ) {                             TypeNode tn = (TypeNode) i.next();
                            print(tn, w, tr);

                            if (i.hasNext()) {
                                    w.write (", ");
                            }
                    }
            }

            w.write(" ");
            print(body(), w, tr);

            w.newline(0);
    }

    public void translate(CodeWriter w, Translator tr) {
        PolyJTypeSystem ts = (PolyJTypeSystem) tr.typeSystem();

        adIntName = PolyJTranslator.adapterInterfaceShortName(this.type);
        PolyJTranslator v = (PolyJTranslator) tr;

        if (type instanceof PolyJClassType) {
            PolyJClassType polyType = (PolyJClassType) type;
            if (ts.needsWhereObject(polyType)) {
                if (type.flags().isInterface()) {
                    v.makeInstanceofClass(polyType);
                } else {
                    v.makeAdapterInterface(polyType);
                }
            }
        }

        if (flags.isInterface()) {
            w.write(flags.clearInterface().clearAbstract().translate());
        } else {
            w.write(flags.translate());
        }

        if (flags.isInterface()) {
            w.write("interface ");
        }
        else {
            w.write("class ");
        }

        w.write(name);

        if ( superClass != null ) {
            w.allowBreak(0, " ");
            w.write( "extends " );
            superClass.del().translate(w, tr);
        }
        if ( !interfaces.isEmpty() ) {
            if (flags.isInterface()) {
                w.allowBreak(0, " ");
                w.write( "extends ");
            }
            else {
                w.allowBreak(0, " ");
                w.write( "implements ");
            }

            for (Iterator iter = interfaces.iterator(); iter.hasNext(); ) {
                TypeNode tn = (TypeNode) iter.next();
                tn.del().translate(w, tr);
                if ( iter.hasNext() ) {
                    w.write( "," );
                    w.allowBreak(0, " ");
                }
            }
        }

        w.write(" { ");
        print(body(), w, tr);
        w.write("}");
        w.newline(0);
    }  

}
