package polyglot.ext.polyj.types;

import polyglot.ast.*;
import polyglot.types.*;
import polyglot.util.*;
import java.util.*;

/**
 * Helper class that encapsulates the logic for constraint
 * checking.
 */
public class ConstraintChecker {

    /**
     * The type system in which this check occurs.
     */
    protected PolyJTypeSystem ts;

    /**
     * The type being instantiated.
     */
    protected PolyJClassType instType;

    /**
     * The arguments on which instType is being instantiated.
     */
    protected List instArgs;

    /**
     * The context in which the instantiation is occuring.
     */
    protected Context instCtxt;

    /**
     * All the errors produced by checking the constraints.
     */
    protected List errors;

    /**
     * Construct a ConstraintChecker.
     * @param instType The polymorphic type being instantiated.
     * @param instArgs The arguments on which instType is being instantiated.
     * @param instantiationContext The context in which the instantiation
     *        occurs.  This determines the visibility and applicability
     *        of methods for the purposes of constraint satisfaction.
     * @param ts The type system in which the instantiation is occuring.
     * @exception SemanticException If the instantiation is illegal.
     */
    public ConstraintChecker(PolyJClassType instType,
            List instArgs, Context instCtxt, PolyJTypeSystem ts) {
        this.instType = instType;
        this.instArgs = instArgs;
        this.instCtxt = instCtxt;
        this.ts = ts;
        errors = new TypedList(new LinkedList(), SemanticException.class, false);
    }

    /**
     * Check whether the arguments provided in the constructor
     * can be used to satisfy the constraints of the polymorphic type.
     */
    public void checkConstraints() throws ConstraintException {
        try {
            ts.checkInstArgs(instType, instArgs);
            List instProxyArgs = convertToProxies(instArgs);
            PolyJTypeWrapper wrapper = ts.Wrapper(instType, instProxyArgs);

            List params = instType.params();
            Iterator parIt = params.iterator();
            Iterator argIt = wrapper.args().iterator();

            while (parIt.hasNext()) {
                ParamType param = (ParamType) parIt.next();
                Type type = (Type) argIt.next();  // this must succeed because of checkInstArgs().
                if (!type.isClass()) {
                    throw new InternalCompilerError("Cannot instantiate on a non-class type", type.position());
                }
                checkCompatibleSignatures(param, type.toClass(), wrapper);
            }
        } catch (SemanticException e) {
            errors.add(e);
        }

        if (!errors.isEmpty()) {
            throw new ConstraintException(errors);
        }
    }

    /**
     * Replace types with their proxies for instantiation checking.
     */
    protected List convertToProxies(List types) throws SemanticException {
        TypedList newTypes = new TypedList(new java.util.ArrayList(), Type.class, false);
        for (Iterator typeIt = types.iterator(); typeIt.hasNext(); ) {
            Type type = (Type)typeIt.next();
            if (type.isPrimitive()) {
                if (type.toPrimitive().isVoid()) {
                    throw new SemanticException(instType.toClass() +
                        " has parameter instantiated with void.");
                } else {
                    type = ts.instantiationProxyType(type.toPrimitive());
                }
            } else if (type.isArray()) {
                type = ts.instantiationProxyType(type.toArray());
            }
            newTypes.add(type);
        }
        return newTypes;
    }
    
    /**
     * Checks whether implementingType has methods that match the
     * signature (including return type and exception types)
     * of constraining type.  In general, constrainingType defines
     * a subset of the methods that implementingType should have.
     * For each method, return types and exception types are allowed
     * to very covariantly between the implementingType and
     * constrainingType.  The argument types are matched using the
     * standard java "best match" algorithm.  It is an error for more
     * than one method to match a set of arguments.  For more
     * details,
     * @see <a href="http://www.pmg.lcs.mit.edu/papers/popl97/node6.html#SECTION00024000000000000000"><em>Satisfying Where Clauses</em> in the PolyJ PoPL paper</a>
     * @param constrainingType a ParamType that has a set of
     *        methods that the implementingType must implement.
     * @param implementingType one of the actual instantiation arguments
     *        that we wish to check for constraint satisfaction
     * @param wrapper a utility object that performs substitutions on
     *        the parameters to the constrained methods
     */
    protected void checkCompatibleSignatures(ParamType constrainingType,
            ClassType implementingType, PolyJTypeWrapper wrapper) {

        Iterator methodIt = constrainingType.methods().iterator();
        while (methodIt.hasNext()) {
            MethodInstance paramMethod = (MethodInstance) methodIt.next();
            MethodInstance want = wrapper.wrapMethod(paramMethod, constrainingType, implementingType, null);
            MethodInstance got;
            try {
                got = ts.findMethod(implementingType, want.name(), want.formalTypes(), instCtxt.currentClass());
            } catch (SemanticException e) {
                recordError(errMsgPrefix(implementingType, constrainingType)
                    + "; " + e.getMessage());
                continue;
            }

            // findMethod() already checked the arguments for us

            // check that the got.returnType() <: want.returnType()
            if (!ts.isSubtype(got.returnType(), want.returnType())) {
                recordError(errMsgPrefix(got, want, implementingType,
                                         constrainingType)
                            + "; return types \"" + got.returnType() +
                            "\" and \"" + want.returnType() +
                            "\" are not covariant");
            }

            // check the exceptions
            checkExceptions(got, want, implementingType, constrainingType);
        }

        Iterator constrIt = constrainingType.constructors().iterator();
        while (constrIt.hasNext()) {
            ConstructorInstance paramConstructor = (ConstructorInstance) constrIt.next();
            ConstructorInstance want = wrapper.wrapConstructor(paramConstructor, constrainingType, implementingType, null);
            ConstructorInstance got;
            try {
                got = ts.findConstructor(implementingType, want.formalTypes(), instCtxt.currentClass());
            } catch (SemanticException e) {
                recordError(errMsgPrefix(implementingType, constrainingType)
                    + "; " + e.getMessage());
                continue;
            }

            // findConstructor() already checked the arguments for us

            // check the exceptions
            checkExceptions(got, want, implementingType, constrainingType);
        }

    }

    /**
     * Verify that the exception got throws are compatible with those
     * that want throws:  for each type g in got.throwTypes(), there is
     * a type w in want.throwTypes() such that g <: w
     */
    protected void checkExceptions(ProcedureInstance got, ProcedureInstance want,
            ClassType implementingType, ParamType constrainingType) {

        // FIXME: This is slow.  Should be faster.
        Iterator gIt = got.throwTypes().iterator();
        while (gIt.hasNext()) {
            Type g = (Type)gIt.next();

            // If the instantiated type declares that it throws
            // some unchecked exception, we can safely
            // skip checking that there's a match for it in the
            // constrained type, because such exceptions are
            // by definition unchecked.
            if (ts.isUncheckedException(g)) {
                continue;
            }

            boolean found = false;
            for (Iterator search = want.throwTypes().iterator();
                 search.hasNext();) {
                Type w = (Type)search.next();
                if (ts.isImplicitCastValid(g, w)) {
                    found = true; break;
                }
            }

            if (!found) {
                recordError(
                            errMsgPrefix(got, want, implementingType, constrainingType)
                            + "; constrained " + want.designator()
                            + " does not throw " + g);
            }
        }
    }

    protected String errMsgPrefix(ProcedureInstance got, ProcedureInstance want,
            ClassType implementingType, ParamType constrainingType) {

        return got.signature() + " in " + implementingType
               + " cannot satisfy "
               + want.signature() + " in " + constrainingType;
    }

    protected String errMsgPrefix(ClassType implementingType, ParamType constrainingType) {

        return implementingType + " cannot satisfy " + constrainingType;
    }

    protected void recordError(String s) {
        errors.add(new SemanticException(s));
    }

    protected void recordError(SemanticException e) {
        errors.add(e);
    }
}

