package polyglot.ext.polyj.types;

import polyglot.ast.*;
import polyglot.frontend.*;
import polyglot.types.*;
import polyglot.util.*;

import polyglot.ext.polyj.types.proxy.*;
import polyglot.ext.polyj.util.*;
import polyglot.ext.polyj.visit.*;

import polyglot.ext.jl.types.*;

import java.util.*;

public class PolyJTypeSystem_c extends TypeSystem_c implements PolyJTypeSystem {

    protected static final Class BOOLEAN_PROXY = BoolProxy.class;
    protected static final Class BYTE_PROXY = ByteProxy.class;
    protected static final Class CHAR_PROXY = CharProxy.class;
    protected static final Class SHORT_PROXY = ShortProxy.class;
    protected static final Class INT_PROXY = IntProxy.class;
    protected static final Class LONG_PROXY = LongProxy.class;
    protected static final Class FLOAT_PROXY = FloatProxy.class;
    protected static final Class DOUBLE_PROXY = DoubleProxy.class;

    protected PolyJTranslationHelper translationHelper;

    // resolver should handle caching.
    public PolyJTypeSystem_c() {
        super();
        translationHelper = new PolyJTranslationHelper(this);
    }


    ///////////////////////////////////////////////////////////////
    //    Overrides of the JL type system
    ///////////////////////////////////////////////////////////////


    public Context createContext() {
        return new PolyJContext_c(this);
    }

/*
    public Resolver classContextResolver(ClassType type) {
        if (type instanceof InstType) {
            return new InstTypeResolver((InstType)type);
        } else {
            return super.classContextResolver(type);
        }
    }
*/

    public ConstructorInstance findConstructor(ClassType clazz, List args, ClassType currClass)
        throws SemanticException
    {
        if (!(clazz instanceof InstType))
            return super.findConstructor(clazz, args, currClass);

        // PolyJ v1 claims the following:
        //   So what we do is we extract the original parametric type that
        //     got instantiated (so the constructor will have a name
        //       like "A.B.[Constructor]" or something like
        //       that, instead of "A.B[int,String,Object].[Constructor]")
        //     and then look it up in the original pjit class.
        // This doesn't seem to work with the way we currently find
        //   constructors, however.  So for now, we'll just call the
        //   superclass implementation and ignore the extracted parametric type.
        InstType instType = (InstType) clazz;
        PolyJClassType paramType = instType.instType();

        return super.findConstructor(clazz, args, currClass);
    }

    /**
     * In PolyJ we have to handle the additional case that we may be
     * calling a (protected/protected/default) method on an instantiation
     * of a polymorphic type from (somewhere, possibly inside an inner
     * class) within that polymorphic type.  Otherwise, behave like the
     * standard type system.
     */
    public boolean isAccessible(MemberInstance mi, Context context)
    {
        ReferenceType target = mi.container();

        if (target instanceof InstType) {
            InstType inst = (InstType) target;
            ClassType c = context.currentClass();
            while (c != null) {
                if (c instanceof PolyJClassType &&
                    isInstantiationOf(inst,(PolyJClassType)c)) {
                    return true;
                }
                c = c.outer();
            }
        }
        return super.isAccessible(mi, context);
    }

    /**
     * In PolyJ, we want to allow assignments to/from proxies and their
     * corresponding primitives.  We also want to disallow assigning
     * null to a parameter type, since the parameter could be instantiated
     * to a primitive.
     */
    public boolean isSubtype(Type child, Type ancestor) {
        if (isProxyAndPrimitive(child, ancestor)) {
            return true;
        }
        if (child.isNull() && isParameter(ancestor)) {
            return false;
        }
        return super.isSubtype(child, ancestor);
    }

    public boolean isImplicitCastValid(Type child, Type ancestor) {
        if (isProxyAndPrimitive(child, ancestor)) {
            return true;
        }
        if (child.isNull() && isParameter(ancestor)) {
            return false;
        }
        return super.isImplicitCastValid(child, ancestor);
    }

    /**
     * In PolyJ, we want to allow implicit casts to/from proxies and their
     * corresponding primitives.
    public boolean isImplicitCastValid(Type from, Type to) {
        return super.isImplicitCastValid(from, to) ||
               isProxyAndPrimitive(from, to);
    }
     */

    /**
     * In PolyJ, we must accept a wider set of casts
     * than in Java.  Specifically, if either fromType
     * or toType is a type parameter, there are instantiations
     * that could allow the cast to succeed.  Thus we have to
     * allow the cast statically so that it will be checked
     * dynamically.
     */
    public boolean isCastValid(Type fromType, Type toType) {
        return super.isCastValid(fromType, toType)
               || isParameter(fromType)
               || isParameter(toType);
    }

    public Set getTypeEncoderRootSet(Type clazz) {
	Set roots = new HashSet();
	    
        // Everything that would be written out in full by the
	// default type system, still should be.  
	// MRC: One reason
	// we seem to need this is that some member classes
	// get represented in uninstantiated form (perhaps in
	// WrappedClasses?).  But I'm not certain that doesn't
	// mean there's a bug in WrappedClasses.
	roots.addAll(super.getTypeEncoderRootSet(clazz));
	    
	// In addition, the null instantiation of clazz should be
	// an additional root, and everything that the default
	// type system would write out for it.  These need to be
	// here for field and method signatures, since these will
	// often have implicit (and thus null) instantiations.
	if (clazz instanceof PolyJClassType && isPolymorphicClass(clazz)) {
	    PolyJClassType pc = (PolyJClassType) clazz;
	    roots.addAll(super.getTypeEncoderRootSet(nullInstantiate(pc)));    
	} 
	
	// Finally, all type parameters should be in the root,
	// since they have no representation as a class file.
	Set params = new HashSet();
        for (Iterator i = roots.iterator(); i.hasNext(); ) {
	    Type t = (Type) i.next();
	    if (isPolymorphicClass(t)) {
		List t_params = allTypeParams(t.toClass());
		params.addAll(t_params);
	    }
	}
	roots.addAll(params);
	
	return roots;
    }

    ///////////////////////////////////////////////////////////////
    //    PolyJ specific
    ///////////////////////////////////////////////////////////////

    /**
     * Return whether c has been instantiated.
     */
    public boolean isInstantiated(ClassType c) {
	return (c instanceof WrappedClassType)
	    || (c instanceof InstType);
    }
    
    /**
     * Return whether t is a type parameter.
     */
    public boolean isParameter(Type t) {
        return t instanceof ParamType;
    }

    /**
     * Return whether t is a polymorphic class type (i.e., a class 
     * type that could be or is instantiated).
     */
    public boolean isPolymorphicClass(Type t) {
        if (t instanceof PolyJClassType) {
            PolyJClassType p = (PolyJClassType) t;
            return p.isPolymorphic();
        }

        if (t instanceof InstType) {
            return true;
        }

        return false;
    }


    /**
     * Return whether p2 is a primitive corresponding
     * to a proxy type that is the same or a superclass of p1.
     */
    protected boolean isProxyAndPrimitive(Type p1, Type p2) {
        if (p1.isPrimitive() && !p2.isPrimitive()) {
            // reverse the arguments
            Type temp = p1;
            p1 = p2;
            p2 = temp;
        }

        // p2 should now be the primitive and p1 the proxy
        if (!p2.isPrimitive() || p1.isPrimitive()) {
            return false;
        }

        PrimitiveType primitive = (PrimitiveType) p2;
        Type proxyType = instantiationProxyType(primitive);
        return isSubtype(proxyType, p1) && equals(p1, Object());
    }

    /**
     * Return true if <m1> and <m2> are the same method.
     */
    public boolean isSameMethod(MethodInstance m1, MethodInstance m2) {
        return m1.name().equals(m2.name()) && hasSameImplArguments(m1, m2);
    }

    /**
     * Return whether p1 and p2 have the same arguments, as they
     * will be implemented in the translation.
     */
    public boolean hasSameImplArguments(ProcedureInstance p1, ProcedureInstance p2) {
        List l1 = p1.formalTypes();
        List l2 = p2.formalTypes();

        Iterator i1 = l1.iterator();
        Iterator i2 = l2.iterator();

        while (i1.hasNext() && i2.hasNext()) {
            Type t1 = (Type) i1.next();
            Type t2 = (Type) i2.next();

            if (! equalsImpl(t1, t2))/*isSameImpl*/ {	
                return false;
            }
        }

        return ! (i1.hasNext() || i2.hasNext());
    }

    /**
     * Return whether t1 and t2 are the same, or whether they
     * are implemented the same in the translation.  That is, a
     * parameter could be a primitive or Object, and a polymorphic type
     * will be Object.
     */
     
     //hackhack try to override equalsImpl(TypeObject t1, TypeObject t2)
    protected boolean equalsImpl(TypeObject t1, TypeObject t2) /*isSameImpl*/ {
        if (equals(t1, t2)) {
            return true;
        }

	if (t1 instanceof Type && t2 instanceof Type)
	{
		Type tt1 = (Type) t1;
		Type tt2 = (Type) t2;
		
		if (tt1.isPrimitive() && isParameter(tt2)
			|| tt2.isPrimitive() && isParameter(tt1)) {
		    return true;
		}
	
		if (isParameter(tt1)) {
		    return equalsImpl(Object(), tt2); /*isSameImpl*/
		}
	
		if (isParameter(tt2)) {
		    return equalsImpl(tt1, Object()); /*isSameImpl*/
		}
	
		if (tt1.isArray() && tt2.isArray()) {
		    Type b1 = tt1.toArray().base();
		    Type b2 = tt2.toArray().base();
		    return equalsImpl(b1, b2);       /*isSameImpl*/
		}
	}
	
        return false;
    }

    public boolean isSameInst(ClassType t1, ClassType t2) {
        ClassType t1_uninst = uninstantiate(t1);
        ClassType t2_uninst = uninstantiate(t2);
        List t1Args = allTypeArgs(t1);
        List t2Args = allTypeArgs(t2);
        return equals(t1_uninst, t2_uninst) && t1Args.equals(t2Args);
    }
    
    /**
     * Return all the type arguments used in forming t, in
     * order from the outermost enclosing type to the innermost.
     * For example, if t is an instance of
     * C[int].D[List[char]], returns [int, List[char]].
     * If t has no such arguments, returns the empty list.
     */
    public List allTypeArgs(ClassType t) {
        List tArgs = null;
        if (t instanceof InstType) {
            InstType i = (InstType) t;
            tArgs = new LinkedList(i.actualArgs());
        } else if (t instanceof WrappedClassType) {
            WrappedClassType w = (WrappedClassType) t;
            tArgs = new LinkedList(w.args());
        } else {
            tArgs = new LinkedList();
        }
        if (t.isInner()) {
            List allArgs = allTypeArgs(t.outer());
            allArgs.addAll(tArgs);
            return allArgs;
        } else {
            return tArgs;
        }
    }

    /**
     * Return all the formal type parameters used in forming t, in
     * order from the outermost enclosing type to the innermost.
     * For example, if t is C[T].D[U], returns [T,U].
     * The elements of the list are ParamTypes.
     * If t has no such arguments, returns the empty list.
     */
    public List allTypeParams(ClassType t) {
        List tParams = null;
        if (t instanceof PolyJClassType) { // also handles WrappedClassTypes
            PolyJClassType p = (PolyJClassType) t;
            tParams = new LinkedList(p.params());
        } else if (t instanceof InstType) {
            InstType i = (InstType) t;
            tParams = new LinkedList(i.instType().params());
        } else {
            tParams = new LinkedList();
        }
        if (t.isInner()) {
            List allParams = allTypeParams(t.outer());
            allParams.addAll(tParams);
            return allParams;
        } else {
            return tParams;
        }
    }

    public String wrapperTypeString(PrimitiveType t) {
        PrimitiveType.Kind k = t.kind();
        String wrapPkg = "polyglot.ext.polyj.wrappers";
        if (k == PrimitiveType.BOOLEAN)
            return wrapPkg + ".BoolWrap";
        if (k == PrimitiveType.CHAR)
            return wrapPkg + ".CharWrap";
        if (k == PrimitiveType.BYTE)
            return wrapPkg + ".ByteWrap";
        if (k == PrimitiveType.SHORT)
            return wrapPkg + ".ShortWrap";
        if (k == PrimitiveType.INT)
            return wrapPkg + ".IntWrap";
        if (k == PrimitiveType.LONG)
            return wrapPkg + ".LongWrap";
        if (k == PrimitiveType.FLOAT)
            return wrapPkg + ".FloatWrap";
        if (k == PrimitiveType.DOUBLE)
            return wrapPkg + ".DoubleWrap";

        throw new InternalCompilerError("Unrecognized primitive type (" + t.getClass().getName() + ")");
    }

    public TranslationTable translationTable() {
        return translationHelper;
    }

    protected static boolean dumped_tables = false;
    public void debug_dumpTranslationTable() {
        if (dumped_tables) return;
        dumped_tables = true;
        System.err.println("Instantiations begin:\n");
        translationHelper.debug_dumpInstantiations();
        System.err.println("Instantiations end.\n\n");

        System.err.println("New Array methods begin:\n");
        translationHelper.debug_dumpNewArrays();
        System.err.println("New Array methods end.\n\n");

        System.err.println("Adapter bindings begin:\n");
        translationHelper.debug_dumpAdapterBindings();
        System.err.println("Adapter bindings end.\n\n");
    }



    /**
     * Returns true iff rt is an instantiation of c
     */
    public boolean isInstantiationOf(InstType rt, PolyJClassType c) {
        return equals(rt.instType(),c);
    }

    /**
     * Check whether the arguments provided in the constructor
     * can be used to satisfy the constraints of the polymorphic type.
     * @param instType The polymorphic type being instantiated.
     * @param instArgs The arguments on which instType is being instantiated.
     * @param instantiationContext The context in which the instantiation
     *        occurs.  This determines the visibility and applicability
     *        of methods for the purposes of constraint satisfaction.
     * @exception SemanticException If the instantiation is illegal.
     */
    public void checkConstraints(PolyJClassType instType,
            List instArgs, Context instCtxt) throws ConstraintException {
        ConstraintChecker cc = new ConstraintChecker(instType, instArgs, instCtxt, this);
        cc.checkConstraints();
    }

    /**
     * Create the type wrapper responsible for instantiating the
     * given polymorphic type t on the argument types.
     */
    public PolyJTypeWrapper Wrapper(PolyJClassType t, List types) {
        return new PolyJTypeWrapper(t, types);
    }

    /**
     * find the method in implementingType that has the
     * signature of paramMethod modulo substitutions done by
     * the wrapper, in the given context.
     */
    public MethodInstance correspondingMethod(MethodInstance paramMethod,
            ParamType constrainingType, ClassType implementingType,
            PolyJTypeWrapper wrapper, Context ctxt)
            throws SemanticException {
        
        MethodInstance want = wrapper.wrapMethod(paramMethod, constrainingType, implementingType, null);
        //System.err.println("trying to match "+want);
        return findMethod(implementingType, want.name(), want.formalTypes(), ctxt.currentClass());
    }

    /**
     * Proxy types double for non-class types during constraint checking
     * of instantiated types.
     */
    public Type instantiationProxyType(PrimitiveType primitive) {
        PrimitiveType.Kind k = primitive.kind();
        try {
            if (k == PrimitiveType.BOOLEAN)
                return typeForClass(BOOLEAN_PROXY);
            if (k == PrimitiveType.CHAR)
                return typeForClass(CHAR_PROXY);
            if (k == PrimitiveType.BYTE)
                return typeForClass(BYTE_PROXY);
            if (k == PrimitiveType.SHORT)
                return typeForClass(SHORT_PROXY);
            if (k == PrimitiveType.INT)
                return typeForClass(INT_PROXY);
            if (k == PrimitiveType.LONG)
                return typeForClass(LONG_PROXY);
            if (k == PrimitiveType.FLOAT)
                return typeForClass(FLOAT_PROXY);
            if (k == PrimitiveType.DOUBLE)
                return typeForClass(DOUBLE_PROXY);
            throw new InternalCompilerError("Unknown kind of primitive: " + k);
        } catch (SemanticException e) {
            throw new InternalCompilerError(e.getMessage());
        }
    }

    /**
     * Proxy types double for non-class types during constraint checking
     * of instantiated types.
     */
    public Type instantiationProxyType(ArrayType ary) {
        Type elemType = ary.base();
        ParsedClassType proxy;

        proxy = createClassType();
        proxy.kind(ClassType.TOP_LEVEL);
        proxy.name("ArrayProxy$" + ary.toString());
        proxy.flags(Public().Abstract().Final());
        proxy.superType(Object());

        // Let A=ary, E=elemType in the following additions:

        // public abstract Object clone();
        proxy.methods().add(methodInstance(Position.COMPILER_GENERATED,
                                           proxy, Public().Abstract(), Object(),
                                           "clone", Collections.EMPTY_LIST, Collections.EMPTY_LIST));

        // public abstract boolean equals(A);
        List equalsArgs = new ArrayList(1);
        equalsArgs.add(proxy);
        proxy.methods().add(methodInstance(Position.COMPILER_GENERATED,
                                           proxy, Public().Abstract(), this.Boolean(),
                                           "equals", equalsArgs, Collections.EMPTY_LIST));

        // public abstract int size();
        proxy.methods().add(methodInstance(Position.COMPILER_GENERATED,
                                           proxy, Public().Abstract(), Int(),
                                           "size", Collections.EMPTY_LIST, Collections.EMPTY_LIST));

        // public abstract E get(int i);
        List getArgs = new ArrayList(1);
        getArgs.add(Int());
        proxy.methods().add(methodInstance(Position.COMPILER_GENERATED,
                                           proxy, Public().Abstract(), elemType,
                                           "get", getArgs, Collections.EMPTY_LIST));

        // public abstract void set(int i, E e);
        List setArgs = new ArrayList(2);
        setArgs.add(Int());
        setArgs.add(elemType);
        proxy.methods().add(methodInstance(Position.COMPILER_GENERATED,
                                           proxy, Public().Abstract(), Void(),
                                           "set", setArgs, Collections.EMPTY_LIST));

        // public A(int s)
        List sizeArgs = new ArrayList(1);
        sizeArgs.add(Int());
        proxy.constructors().add(constructorInstance(Position.COMPILER_GENERATED,
                                                     proxy, Public(), sizeArgs, Collections.EMPTY_LIST));

        // public A(A)
        List copyArgs = new ArrayList(1);
        copyArgs.add(proxy);
        proxy.constructors().add(constructorInstance(Position.COMPILER_GENERATED,
                                                     proxy, Public(), copyArgs, Collections.EMPTY_LIST));

        // public A(E[])
        List arrayArgs = new ArrayList(1);
        arrayArgs.add(arrayOf(elemType));
        proxy.constructors().add(constructorInstance(Position.COMPILER_GENERATED,
                                                     proxy, Public(), arrayArgs, Collections.EMPTY_LIST));

        return proxy;
    }

    /**
     * Return the uninstantiated type of c.
     */
    public PolyJClassType uninstantiate(ClassType c) {
        PolyJClassType p = null;

        if (c instanceof WrappedClassType) {
            p = uninstantiate(((WrappedClassType) c).wrappedClass());
        } else if (c instanceof InstType) {
            p = uninstantiate(((InstType) c).instType());
        } else if (c instanceof PolyJClassType) {
            p = (PolyJClassType) c;
        } else {
            throw new InternalCompilerError("Unexpected class type");
        }

        if (allTypeArgs(p).size() != 0) {
            //return uninstantiate(c);
            throw new InternalCompilerError(p + " isn't fully uninstantiated");
        }

        return p;
    }

    /**
     * Instantiate t and its outer class on their own parameters.
     * @requires t or one of its outer classes must be parameterized
     */
    public ClassType nullInstantiate(PolyJClassType t) {
	if (allTypeParams(t).isEmpty()) {
	    throw new InternalCompilerError("Cannot null instantiate "
	        + "unparameterized class " + t);
	}

	return uncheckedNullInstantiate(t);
    }

    /**
     * Instantiate t and its outer class on their own parameters.
     * @requires t is a member or top-level class.
     */
    protected ClassType uncheckedNullInstantiate(ClassType t) {
	if (t.isInner()) {
	    ClassType outer_ni = uncheckedNullInstantiate(t.outer());

	    if (t.isMember()) {
		// Replace t with its equivalent from the now null instantiated
		// outer class.
	        t = outer_ni.memberClassNamed(t.name());
	    } else {
		// MRC: The problem here is: how do we find t in outer_ni if
		// it has no name?  There are no methods in ClassType
		// to do this.  We could create them, e.g., 
		// ClassType.findUid(), but we may also want to eliminate
		// locals/anons/etc later.  So for now I'm leaving this
		// partially defined.
		throw new InternalCompilerError("Null instantiation is " +
		    "unimplemented for local and anonymous classes");
	    }
	} 

	if (t instanceof PolyJClassType) {
	    PolyJClassType p = (PolyJClassType) t;
	    if (!p.params().isEmpty()) {
    	        try {
		    // Replace t with its immediate null instantiation.
                    t = instantiate(p, p.params());
                } catch (SemanticException e) {
                    throw new InternalCompilerError("Null instantiation "
		        + "failed: " + e.getMessage());
                }
	    }
	} 
	    
	return t;
    }
    
    /**
     * Instantiate the parameterized type <code>t</code> on <code>args</code>.
     */
    public InstType instantiate(PolyJClassType t, List args) throws SemanticException {
        return instantiate(Position.COMPILER_GENERATED, t, args);
    }

    /**
     * Instantiate the parameterized type <code>t</code> on <code>args</code>.
     */
    public InstType instantiate(Position pos, PolyJClassType t, List args) throws SemanticException {
        checkInstArgs(t, args);
        InstType theResult = uncheckedInstantiate(t, args);
        addInstantiation(theResult);
        return theResult;
    }

    /**
     * Instantiate t on args, but some checks.
     */
    public InstType uncheckedInstantiate(PolyJClassType t, List args) {
        return uncheckedInstantiate(Position.COMPILER_GENERATED, t, args);
    }

    /**
     * Instantiate t on args at pos, but skip some checks.
     */
    public InstType uncheckedInstantiate(Position pos, PolyJClassType t, List args) {
        return new InstType_c(pos, t, args, this);
    }
    
    /**
     * Check that the we have the correct number and type of arguments.
     */
    public void checkInstArgs(PolyJClassType t, List args) throws SemanticException {
        List params = t.params();
        int nParams = params.size();
        int nArgs = args.size();

        if (nArgs < nParams) {
            throw new SemanticException ("Too few parameters in instantiation of " +
                t.toClass() + " (" + nArgs + " but should be " + nParams + ")");
        } else if (nArgs > nParams) {
            throw new SemanticException ("Too many parameters in instantiation of " +
                t.toClass() + " (" + nArgs + " but should be " + nParams + ")");
        }

        Iterator parIt = params.iterator();
        Iterator argIt = args.iterator();

        while (parIt.hasNext()) {
            ParamType param = (ParamType)parIt.next();
            Object o = argIt.next();
            if (o instanceof Type) {
                // Everything is fine.
            } else if (o instanceof Expr) {
                Expr e = (Expr) o;
                throw new SemanticException("Value parameters are not supported.", e.position());
            } else if (o instanceof Node) {
                Node n = (Node) o;
                throw new SemanticException("Instantiation parameter must be a type or value.", n.position());
            } else {
                throw new InternalCompilerError("Unknown kind of instantiation parameter.", t.position());
            }
        }
    }

    /**
     * Get the method in the original (completely uninstantiated) type
     * that corresponds to this method instance.
     */
    public MethodInstance originalMethod(MethodInstance mi) {
        Type et = null;

        while (mi != null && ((et = mi.container()) instanceof BackMapProcedures)) {
            MethodInstance back = ((BackMapProcedures)et).originalMethod(mi);
            if (back == null) {
                return mi;
            } else {
                mi = back;
            }
        }

        return mi;
    }

    /**
     * Get the constructor in the original (completely uninstantiated) type
     * that corresponds to this constructor instance.
     */
    public ConstructorInstance originalConstructor(ConstructorInstance ci) {
        Type et = null;

        while (ci != null && ((et = ci.container()) instanceof BackMapProcedures)) {
            ConstructorInstance back = ((BackMapProcedures)et).originalConstructor(ci);
            if (back == null) {
                return ci;
            } else {
                ci = back;
            }
        }

        return ci;
    }

    /**
     * Find the original FieldInstance in <code>c</code> that defines
     * field <code>fi</code>.
     */
    public FieldInstance originalField(ClassType c, FieldInstance fi) {
        for (Iterator i = c.fields().iterator(); i.hasNext(); ) {
            FieldInstance curFI = (FieldInstance) i.next();
            if (curFI.name().equals(fi.name())) {
                return curFI;
            }
        }

        throw new InternalCompilerError(fi.position(),
            "No FieldInstance could be found for this field");
    }
   
    /**
     * Record in the translator that c is a type that was
     * instantiated in the source.
     */
    public void addInstantiation(ClassType c) {
        translationHelper.addInstantiation(c);
    }

    public void addInstantiations(Collection types) {
        for (Iterator i = types.iterator(); i.hasNext(); ) {
            addInstantiation((ClassType)i.next());
        }
    }

    public void addNewArrayMethod(ParamType param, int dims) {
        translationHelper.addNewArrayMethod(param, dims);
    }

    /**
     * Assisted instantiations are situations where a polymorphic type
     * is instantiating another polymorphic type with one or more of
     * its own parameter type.  We need a method in the where object
     * to return a where object for the instance class.
     */
    public void addAssistedInstantiation(PolyJClassType assistant,
        ClassType instance)
    {
        translationHelper.addAssistedInstantiation(assistant, instance);
    }

    /**
     * Return whether any of <code>c</code>'s enclosing instances
     * have a where object field.
     */
    public boolean hasOuterWhereObject(ClassType c) {
	if (c.isInner()) {
	    ClassType o = c.outer();
	    if (needsWhereObject(o)) {
 	        return true;
	    }
	}

	return false;
    }
    
    /**
     * Returns true iff some class up the type hierarchy from t
     *   has a where-object field.
     *
     * A class inherits a where-object if its super-type needs a
     * where object (which is to say, its super-type has a where object).
     */
    public boolean inheritsWhereObject(ReferenceType t) {
        if (t.superType() != null) {
            if (!t.superType().isReference())
                throw new InternalCompilerError("Supertype of class "+ t +
                    " is not a reference type");
            if (needsWhereObject(t.superType().toReference()))
                return true;
        }
        return false;
    }

    /**
     * A class needs a where-object if its an instantiated type or
     * a polymorphic type, or if its supertype or one of its super-interfaces
     * needs a where object, or if its enclosing class needs a where
     * object.
     */
    public boolean needsWhereObject(ReferenceType t) {
        if (t instanceof InstType ||
               t instanceof PolyJClassType &&
               ((PolyJClassType)t).isPolymorphic()) {
            return true;
        }

        if (t.superType() != null) {
            if (!t.superType().isReference()) {
                throw new InternalCompilerError("Supertype of class "+ t +
                    " is not a reference type");
            }
            if (needsWhereObject(t.superType().toReference())) {
                return true;
            }
        }

        for (Iterator it = t.interfaces().iterator(); it.hasNext(); ) {
            Type i = (Type) it.next();
            if (!i.isReference()) {
                throw new InternalCompilerError("Super-interface " + i +
                    " of class " + t +
                    " is not a reference type");
            }
            if (needsWhereObject(i.toReference())) {
                return true;
            }
        }

        if (t.isClass() && t.toClass().isInner()) {
            if (needsWhereObject(t.toClass().outer())) {
                return true;
            }
        }

        return false;
    }

    /**
     * Return true iff t is a type that should have an
     * adapter class generated for it.
     */
    public boolean needsAdapterClass(Type t) {
        if (!t.isClass()) {
            return false;
        }

        ClassType c = t.toClass();

        if (!needsWhereObject(c)) {
            return false;
        }

        if (!isFullyInst(c)) {
            return false;
        }

        return true;
    }

    /**
     * Return whether all the type parameters of <code>t</code>
     * have been instantiated.
     */
    public boolean allParamsInst(ClassType t) {
        return allTypeParams(t).size() == allTypeArgs(t).size();
    }
    
    /**
     * Return whether t has been fully instantiated on
     * all its type parameters.  That is, whether the
     * structure of t includes any type parameters.
     */
    public boolean isFullyInst(ClassType t) {
        if (!allParamsInst(t)) {
            return false;
        }
        
        for (Iterator i = allTypeArgs(t).iterator(); i.hasNext(); ) {
            Type p = (Type) i.next();
            if (p instanceof ParamType 
                || p.isClass() && !isFullyInst(p.toClass())) 
            {
                return false;
            }
        }
        
        return true;
    }

    /**
     * A prmitive type wrapper is the runtime representation of a
     * primitive type when it is manipulated by a polymorphic type.
     */
    public String primitiveTypeWrapperName(PrimitiveType primitive) {
        String PACKAGE="polyglot.ext.polyj.wrappers";
        PrimitiveType.Kind k = primitive.kind();

        if (k == PrimitiveType.BOOLEAN)
            return PACKAGE + ".BoolWrap";
        if (k == PrimitiveType.CHAR)
            return PACKAGE + ".CharWrap";
        if (k == PrimitiveType.BYTE)
            return PACKAGE + ".ByteWrap";
        if (k == PrimitiveType.SHORT)
            return PACKAGE + ".ShortWrap";
        if (k == PrimitiveType.INT)
            return PACKAGE + ".IntWrap";
        if (k == PrimitiveType.LONG)
            return PACKAGE + ".LongWrap";
        if (k == PrimitiveType.FLOAT)
            return PACKAGE + ".FloatWrap";
        if (k == PrimitiveType.DOUBLE)
            return PACKAGE + ".DoubleWrap";

        throw new InternalCompilerError("called with a primitive with kind=" + k);
    }

    /**
     * Returns t, modified as necessary to make it a legal
     * static target.
     */
    public Type staticTarget(Type t) {
        // In PolyJ, creating a static target means we have to
        // add any missing type parameters to the target.
        // Otherwise, CanonicalTypeNode will complain.

	if (t.isClass() && t.toClass().isMember()) {
	    ClassType m = t.toClass();
	    t = staticTarget(m.outer()).toClass().memberClassNamed(m.name());
	}
	    
        if (t instanceof PolyJClassType) {
            PolyJClassType ct = (PolyJClassType) t;
            // NJN: removed ct.isPolymorphic() check.  The check does not
            // work when ct is an inner class with no parameters--we should
            // null instantiate its container.  In general, polyj is very
            // bad about distinguishing between PolyJClassType and
            // its null instantiation.  We should not have them be considered
            // equal as is done now.
            if (ct.isPolymorphic() && !isInstantiated(ct)) {
                t = nullInstantiate(ct);
            }
        }

        return t;
    }

    // Type Factories

    protected ArrayType arrayType(Position pos, Type base) {
	return new PolyJArrayType_c(this, pos, base);
    }
    
    public ParamType ParamType(Position pos, PolyJClassType constrainedClass,
            String paramName, int index) {
        return new ParamType_c(this, pos, constrainedClass, paramName, index);
    }

    public ParsedClassType createClassType(LazyClassInitializer init, 
                                           Source fromSource) {
        if (! init.fromClassFile()) {
            return new PolyJParsedClassType_c(this, init, fromSource);
        }
        else {
            return super.createClassType(init, fromSource);
        }
    }

    public PolyDlg PolyDlg(PolyJClassType ct) {
        return new PolyDlg_c(ct);
    }

    public WrapDlg WrapDlg(WrappedClassType w,
            PolyJClassType r, ClassType oldEnclosingType,
            ClassType newEnclosingType, PolyJTypeWrapper tw) {
        return new WrapDlg_c(w, r, oldEnclosingType, newEnclosingType, tw);
    }

    public WrappedClassType WrappedClassType(PolyJClassType r,
            ClassType oldEnclosingType, ClassType newEnclosingType,
            PolyJTypeWrapper tw) {
	return r.wrap(oldEnclosingType, newEnclosingType, tw);
    }

    public ArrayAccessBinding ArrayAccessBinding(ParamType param, 
        ClassType enclosing) 
    {
        return new ArrayAccessBinding_c(param, enclosing, this);
    }

    public InstantiationBinding InstantiationBinding(ClassType c) {
        return new InstantiationBinding_c(c);
    }

    public NewArrayBinding NewArrayBinding(ParamType param, int dims,
        ClassType enclosing) 
    {
        return new NewArrayBinding_c(param, dims, enclosing, this);
    }

    public AdapterBinding AdapterBinding(PolyJClassType assistant,
        ClassType instantiation)
    {
        return new AdapterBinding_c(assistant, instantiation);
    }

    public ParamSigBinding ParamSigBinding(ParamType param,
        ClassType enclosing)
    {
        return new ParamSigBinding_c(param, enclosing, this);
    }

    public DefaultBinding DefaultBinding(ParamType param, 
        ClassType enclosing)
    {
        return new DefaultBinding_c(param, enclosing, this);
    }

    public WhereBinding WhereBinding(ParamType param, 
	ProcedureInstance constraint)
    {
	return new WhereBinding_c(param, constraint);
    }

    public StaticFieldBinding StaticFieldBinding(ClassType enclosingType, FieldInstance fi)
    {
        return new StaticFieldBinding_c(enclosingType, fi);
    }
}
